# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
from copy import deepcopy


class VisElementProperties(t.TypedDict):
    name: str
    type: str
    doc: str
    default_value: t.Any
    default_property: t.Any


class VisElementDetail(t.TypedDict):
    inherits: t.List[str]
    properties: t.List[VisElementProperties]


VisElement = t.Tuple[str, VisElementDetail]


class VisElements(t.TypedDict):
    blocks: t.List[VisElement]
    controls: t.List[VisElement]
    undocumented: t.List[VisElement]


def _resolve_inherit_property(element: VisElement, viselements: VisElements) -> t.List[VisElementProperties]:
    element_detail = element[1]
    properties = deepcopy(element_detail["properties"])
    if "inherits" not in element_detail:
        return properties
    for inherit in element_detail["inherits"]:
        inherit_element = None
        for element_type in "blocks", "controls", "undocumented":
            inherit_element = next((e for e in viselements[element_type] if e[0] == inherit), None)  # type: ignore[literal-required]
            if inherit_element is not None:
                break
        if inherit_element is None:
            raise RuntimeError(f"Error resolving inherit element with name {inherit} in viselements.json")
        properties = properties + _resolve_inherit_property(inherit_element, viselements)
    return properties


def resolve_inherits(viselements: VisElements) -> VisElements:
    """NOT DOCUMENTED"""
    for element_type in "blocks", "controls":
        for element in viselements[element_type]:  # type: ignore[literal-required]
            element[1]["properties"] = _resolve_inherit_property(element, viselements)
    return viselements



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t

import pandas as pd

from .._warnings import _warn
from ..gui import Gui
from ..utils import _getscopeattr


def _compare_function(
    gui: Gui, compare_name: str, name: str, value: pd.DataFrame, datanames: str
) -> t.Optional[pd.DataFrame]:
    try:
        names = datanames.split(",")
        if not names:
            return None
        compare_fn = gui._get_user_function(compare_name) if compare_name else None
        if callable(compare_fn):
            return gui._get_accessor().to_pandas(
                gui._call_function_with_state(compare_fn, [name, [gui._get_real_var_name(n) for n in names]])
            )
        elif compare_fn is not None:
            _warn(f"{compare_name}(): compare function name is not valid.")
        dfs = [gui._get_accessor().to_pandas(_getscopeattr(gui, n)) for n in names]
        return value.compare(dfs[0], keep_shape=True)
    except Exception as e:
        if not gui._call_on_exception(compare_name or "Gui._compare_function", e):
            _warn(f"{compare_name or 'Gui._compare_function'}(): compare function raised an exception", e)
        return None



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t


class Icon:
    """Small image in the User Interface.

    Icons are typically used in controls like
    [button](../../../../../refmans/gui/viselements/generic/button.md)
    or items in a [menu](../../../../../refmans/gui/viselements/generic/menu.md).

    If a text is associated to an icon, it is rendered by the visual elements that
    uses this Icon.
    """

    @staticmethod
    def get_dict_or(value: t.Union[str, t.Any]) -> t.Union[str, dict]:
        return value._to_dict() if isinstance(value, Icon) else value

    def __init__(
        self,
        path: str,
        text: t.Optional[str] = None,
        light_path: t.Optional[str] = None,
        dark_path: t.Optional[str] = None,
    ) -> None:
        """Initialize a new Icon.

        Arguments:
            path (str): The path to an image file.<br/>
                This path must be relative to and inside the root directory of the
                application (where the Python file that created the `Gui` instance
                is located).<br/>
                In situations where the image file is located outside the application
                directory, a _path mapping_ must be used: the _path_mapping_ parameter
                of the `Gui^` constructor makes it possible to access a resource anywhere
                on the server filesystem, as if it were located in a subdirectory of
                the application directory.
            text (Optional[str]): The text associated to the image. If *text* is None,
                there is no text associated to this image.
            light_path (Optional[str]): The path to the light theme image (fallback to *path* if not defined).
            dark_path (Optional[str]): The path to the dark theme image (fallback to *path* if not defined).
        """
        self.path: str = path
        self.text: t.Optional[str] = text
        if light_path is not None:
            self.light_path: t.Optional[str] = light_path
        if dark_path is not None:
            self.dark_path: t.Optional[str] = dark_path

    def _to_dict(self, a_dict: t.Optional[dict] = None) -> dict:
        if a_dict is None:
            a_dict = {}
        a_dict["path"] = self.path
        if self.text is not None:
            a_dict["text"] = self.text
        if hasattr(self, "light_path") and self.light_path is not None:
            a_dict["lightPath"] = self.light_path
        if hasattr(self, "dark_path") and self.dark_path is not None:
            a_dict["darkPath"] = self.dark_path
        return a_dict



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from importlib import util


def _is_in_notebook():  # pragma: no cover
    try:
        if not util.find_spec("IPython"):
            return False

        from IPython import get_ipython  # type: ignore[reportPrivateImportUsage]

        ipython = get_ipython()

        if ipython is None or "IPKernelApp" not in ipython.config:
            return False
    except (ImportError, AttributeError):
        return False
    return True



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import ast
import builtins
import typing as t
from operator import attrgetter
from types import FrameType

_python_builtins = dir(builtins)


def _get_value_in_frame(frame: FrameType, name: str):
    if not frame:
        return None
    if name in frame.f_locals:
        return frame.f_locals.get(name)
    return _get_value_in_frame(t.cast(FrameType, frame.f_back), name)


class _TransformVarToValue(ast.NodeTransformer):
    def __init__(self, frame: FrameType, non_vars: t.List[str]) -> None:
        super().__init__()
        self.frame = frame
        self.non_vars = non_vars

    def visit_Name(self, node):
        var_parts = node.id.split(".", 2)
        if var_parts[0] in self.non_vars:
            return node
        value = _get_value_in_frame(self.frame, var_parts[0])
        if len(var_parts) > 1:
            value = attrgetter(var_parts[1])(value)
        if not isinstance(value, (str, int, float, bool, list, tuple)):
            # transform into constants only what can be (ie not callable or generator for example)
            return node
        return ast.Constant(value=value, kind=None)


class _LambdaByName(ast.NodeVisitor):
    _DEFAULT_NAME = "<default>"

    def __init__(self, element_name: str, lineno: int, lambdas: t.Dict[str, ast.Lambda]) -> None:
        super().__init__()
        self.element_name = element_name.split(".")[-1]
        self.lambdas = lambdas
        self.lineno = lineno + 1

    def visit_Call(self, node):
        if getattr(node.func, "attr", None) == self.element_name:
            if self.lambdas.get(_LambdaByName._DEFAULT_NAME, None) is None and (
                a_lambda := next(
                    (
                        arg
                        for arg in node.args
                        if isinstance(arg, ast.Lambda)
                        and arg.lineno is not None
                        and arg.end_lineno is not None
                        and self.lineno >= arg.lineno
                        and self.lineno <= arg.end_lineno
                    ),
                    None,
                )
            ):
                self.lambdas[_LambdaByName._DEFAULT_NAME] = a_lambda

            for kwd in node.keywords:
                if (
                    kwd.arg is not None
                    and isinstance(kwd.value, ast.Lambda)
                    and kwd.value.lineno is not None
                    and kwd.value.end_lineno is not None
                    and self.lineno >= kwd.value.lineno
                    and self.lineno <= kwd.value.end_lineno
                ):
                    self.lambdas[kwd.arg] = kwd.value



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Dict, Tuple

from taipy.common._cli._base_cli._abstract_cli import _AbstractCLI
from taipy.common._cli._base_cli._taipy_parser import _TaipyParser

from ._hook import _Hooks


class _GuiCLI(_AbstractCLI):
    """Command-line interface of GUI."""

    __GUI_ARGS: Dict[Tuple, Dict] = {
        ("--port", "-P"): {
            "dest": "taipy_port",
            "metavar": "PORT",
            "nargs": "?",
            "default": "",
            "const": "",
            "help": "Specify server port",
        },
        ("--host", "-H"): {
            "dest": "taipy_host",
            "metavar": "HOST",
            "nargs": "?",
            "default": "",
            "const": "",
            "help": "Specify server host",
        },
        ("--client-url",): {
            "dest": "taipy_client_url",
            "metavar": "CLIENT_URL",
            "nargs": "?",
            "default": "",
            "const": "",
            "help": "Specify client URL",
        },
        ("--ngrok-token",): {
            "dest": "taipy_ngrok_token",
            "metavar": "NGROK_TOKEN",
            "nargs": "?",
            "default": "",
            "const": "",
            "help": "Specify NGROK Authtoken",
        },
        ("--webapp-path",): {
            "dest": "taipy_webapp_path",
            "metavar": "WEBAPP_PATH",
            "nargs": "?",
            "default": "",
            "const": "",
            "help": "The path to the web app to be used. The default is the webapp directory under gui in the Taipy GUI package directory.",  # noqa: E501
        },
        ("--upload-folder",): {
            "dest": "taipy_upload_folder",
            "metavar": "UPLOAD_FOLDER",
            "nargs": "?",
            "default": "",
            "const": "",
            "help": "The path to the folder where uploaded files from Taipy GUI will be stored.",
        },
    }

    __DEBUG_ARGS: Dict[str, Dict] = {
        "--debug": {"dest": "taipy_debug", "help": "Turn on debug", "action": "store_true"},
        "--no-debug": {"dest": "taipy_no_debug", "help": "Turn off debug", "action": "store_true"},
    }

    __RELOADER_ARGS: Dict[str, Dict] = {
        "--use-reloader": {"dest": "taipy_use_reloader", "help": "Auto reload on code changes", "action": "store_true"},
        "--no-reloader": {"dest": "taipy_no_reloader", "help": "No reload on code changes", "action": "store_true"},
    }

    __BROWSER_ARGS: Dict[str, Dict] = {
        "--run-browser": {
            "dest": "taipy_run_browser",
            "help": "Open a new tab in the system browser",
            "action": "store_true",
        },
        "--no-run-browser": {
            "dest": "taipy_no_run_browser",
            "help": "Don't open a new tab for the application",
            "action": "store_true",
        },
    }

    __DARK_LIGHT_MODE_ARGS: Dict[str, Dict] = {
        "--dark-mode": {
            "dest": "taipy_dark_mode",
            "help": "Apply dark mode to the GUI application",
            "action": "store_true",
        },
        "--light-mode": {
            "dest": "taipy_light_mode",
            "help": "Apply light mode to the GUI application",
            "action": "store_true",
        },
    }

    @classmethod
    def create_parser(cls):
        gui_parser = _TaipyParser._add_groupparser("Taipy GUI", "Optional arguments for Taipy GUI service")

        for args, arg_dict in cls.__GUI_ARGS.items():
            arg = (args[0], cls.__add_taipy_prefix(args[0]), *args[1:])
            gui_parser.add_argument(*arg, **arg_dict)

        debug_group = gui_parser.add_mutually_exclusive_group()
        for arg, arg_dict in cls.__DEBUG_ARGS.items():
            debug_group.add_argument(arg, cls.__add_taipy_prefix(arg), **arg_dict)

        reloader_group = gui_parser.add_mutually_exclusive_group()
        for arg, arg_dict in cls.__RELOADER_ARGS.items():
            reloader_group.add_argument(arg, cls.__add_taipy_prefix(arg), **arg_dict)

        browser_group = gui_parser.add_mutually_exclusive_group()
        for arg, arg_dict in cls.__BROWSER_ARGS.items():
            browser_group.add_argument(arg, cls.__add_taipy_prefix(arg), **arg_dict)

        dark_light_mode_group = gui_parser.add_mutually_exclusive_group()
        for arg, arg_dict in cls.__DARK_LIGHT_MODE_ARGS.items():
            dark_light_mode_group.add_argument(arg, cls.__add_taipy_prefix(arg), **arg_dict)

        if (hook_cli_arg := _Hooks()._get_cli_args()) is not None:
            hook_group = gui_parser.add_mutually_exclusive_group()
            for hook_arg, hook_arg_dict in hook_cli_arg.items():
                hook_group.add_argument(hook_arg, cls.__add_taipy_prefix(hook_arg), **hook_arg_dict)

    @classmethod
    def create_run_parser(cls):
        run_parser = _TaipyParser._add_subparser("run", help="Run a Taipy application.")
        for args, arg_dict in cls.__GUI_ARGS.items():
            run_parser.add_argument(*args, **arg_dict)

        debug_group = run_parser.add_mutually_exclusive_group()
        for arg, arg_dict in cls.__DEBUG_ARGS.items():
            debug_group.add_argument(arg, **arg_dict)

        reloader_group = run_parser.add_mutually_exclusive_group()
        for arg, arg_dict in cls.__RELOADER_ARGS.items():
            reloader_group.add_argument(arg, **arg_dict)

        browser_group = run_parser.add_mutually_exclusive_group()
        for arg, arg_dict in cls.__BROWSER_ARGS.items():
            browser_group.add_argument(arg, **arg_dict)

        dark_light_mode_group = run_parser.add_mutually_exclusive_group()
        for arg, arg_dict in cls.__DARK_LIGHT_MODE_ARGS.items():
            dark_light_mode_group.add_argument(arg, **arg_dict)

        if (hook_cli_arg := _Hooks()._get_cli_args()) is not None:
            hook_group = run_parser.add_mutually_exclusive_group()
            for hook_arg, hook_arg_dict in hook_cli_arg.items():
                hook_group.add_argument(hook_arg, **hook_arg_dict)

    @classmethod
    def handle_command(cls):
        args, _ = _TaipyParser._parser.parse_known_args()
        return args

    @classmethod
    def __add_taipy_prefix(cls, key: str):
        if key.startswith("--no-"):
            return key[:5] + "taipy-" + key[5:]

        return key[:2] + "taipy-" + key[2:]



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t

from taipy.common.logger._taipy_logger import _TaipyLogger

from .utils.singleton import _Singleton


class _Hook:
    method_names: t.List[str] = []


class _Hooks(object, metaclass=_Singleton):
    def __init__(self):
        self.__hooks: t.List[_Hook] = []  # type: ignore[annotation-unchecked]

    def _register_hook(self, hook: _Hook):
        # Prevent duplicated hooks
        for h in self.__hooks:
            if type(hook) is type(h):
                _TaipyLogger._get_logger().info(f"Failed to register duplicated hook of type '{type(h)}'")
                return
        self.__hooks.append(hook)

    def __getattr__(self, name: str):
        def _resolve_hook(*args, **kwargs):
            for hook in self.__hooks:
                if name not in hook.method_names:
                    continue
                # call hook
                try:
                    func = getattr(hook, name)
                    if not callable(func):
                        raise Exception(f"'{name}' hook is not callable")
                    res = getattr(hook, name)(*args, **kwargs)
                except Exception as e:
                    _TaipyLogger._get_logger().error(f"Error while calling hook '{name}': {e}")
                    return
                # check if the hook returns True -> stop the chain
                if res is True:
                    return
                # only hooks that return true are allowed to return values to ensure consistent response
                if isinstance(res, (list, tuple)) and len(res) == 2 and res[0] is True:
                    return res[1]

        return _resolve_hook



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t

import numpy
import pandas as pd

from .pandas_data_accessor import _PandasDataAccessor


class _NumpyDataAccessor(_PandasDataAccessor):
    __types = (numpy.ndarray,)

    @staticmethod
    def get_supported_classes() -> t.List[t.Type]:
        return list(_NumpyDataAccessor.__types)

    def to_pandas(self, value: t.Any) -> pd.DataFrame:
        return pd.DataFrame(value)

    def _from_pandas(self, value: pd.DataFrame, type: t.Type):
        if type is numpy.ndarray:
            return value.to_numpy()
        return super()._from_pandas(value, type)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
import typing as t

from .._warnings import _warn
from .boolean import _is_boolean, _is_true
from .clientvarname import _to_camel_case


def _get_column_desc(columns: t.Dict[str, t.Any], key: str) -> t.Optional[t.Dict[str, t.Any]]:
    return next((x for x in columns.values() if x.get("dfid") == key), None)


def _get_name_indexed_property(attributes: t.Dict[str, t.Any], name: str) -> t.Dict[str, t.Any]:
    ret = {}
    index_re = re.compile(name + r"\[(.*)\]$")
    for key in attributes.keys():
        if m := index_re.match(key):
            ret[m.group(1)] = attributes.get(key)
    return ret


def _update_col_desc_from_indexed(
    attributes: t.Dict[str, t.Any], columns: t.Dict[str, t.Any], name: str, elt_name: str
):
    col_value = _get_name_indexed_property(attributes, name)
    for k, v in col_value.items():
        if col_desc := next((x for x in columns.values() if x.get("dfid") == k), None):
            if col_desc.get(_to_camel_case(name)) is None:
                col_desc[_to_camel_case(name)] = str(v)
        else:
            _warn(f"{elt_name}: '{k}' is not a displayed column in {name}[].")


def _enhance_columns(  # noqa: C901
    attributes: t.Dict[str, t.Any], hash_names: t.Dict[str, str], columns: t.Dict[str, t.Any], elt_name: str
):
    _update_col_desc_from_indexed(attributes, columns, "nan_value", elt_name)
    _update_col_desc_from_indexed(attributes, columns, "width", elt_name)
    filters = _get_name_indexed_property(attributes, "filter")
    for k, v in filters.items():
        if _is_true(v):
            if col_desc := _get_column_desc(columns, k):
                col_desc["filter"] = True
            else:
                _warn(f"{elt_name}: '{k}' is not a displayed column for filter[].")
    editables = _get_name_indexed_property(attributes, "editable")
    for k, v in editables.items():
        if _is_boolean(v):
            if col_desc := _get_column_desc(columns, k):
                col_desc["notEditable"] = not _is_true(v)
            else:
                _warn(f"{elt_name}: '{k}' is not a displayed column in editable[].")
    group_by = _get_name_indexed_property(attributes, "group_by")
    for k, v in group_by.items():
        if _is_true(v):
            if col_desc := _get_column_desc(columns, k):
                col_desc["groupBy"] = True
            else:
                _warn(f"{elt_name}: '{k}' is not a displayed column in group_by[].")
    apply = _get_name_indexed_property(attributes, "apply")
    for k, v in apply.items():  # pragma: no cover
        if col_desc := _get_column_desc(columns, k):
            if callable(v):
                value = hash_names.get(f"apply[{k}]")
            elif isinstance(v, str):
                value = v.strip()
            else:
                _warn(f"{elt_name}: invalid user or predefined function in apply[].")
                value = None
            if value:
                col_desc["apply"] = value
        else:
            _warn(f"{elt_name}: '{k}' is not a displayed column in apply[].")
    if _get_name_indexed_property(attributes, "style"):
        _warn("Table: property 'style[]' has been renamed to 'cell_class_name[]'.")
    cell_class_names = _get_name_indexed_property(attributes, "cell_class_name")
    for k, v in cell_class_names.items():  # pragma: no cover
        if col_desc := _get_column_desc(columns, k):
            if callable(v):
                value = hash_names.get(f"cell_class_name[{k}]")
            elif isinstance(v, str):
                value = v.strip()
            else:
                value = None
            if value in columns.keys():
                _warn(f"{elt_name}: cell_class_name[{k}] cannot be set to an column name '{value}'.")
            elif value:
                col_desc["className"] = value
        else:
            _warn(f"{elt_name}: '{k}' is not a displayed column in cell_class_name[].")
    tooltips = _get_name_indexed_property(attributes, "tooltip")
    for k, v in tooltips.items():  # pragma: no cover
        if col_desc := _get_column_desc(columns, k):
            if callable(v):
                value = hash_names.get(f"tooltip[{k}]")
            elif isinstance(v, str):
                value = v.strip()
            else:
                value = None
            if value in columns.keys():
                _warn(f"{elt_name}: tooltip[{k}] cannot be set to an column name '{value}'.")
            elif value:
                col_desc["tooltip"] = value
        else:
            _warn(f"{elt_name}: '{k}' is not a displayed column in tooltip[].")
    formats = _get_name_indexed_property(attributes, "format_fn")
    for k, v in formats.items():  # pragma: no cover
        if col_desc := _get_column_desc(columns, k):
            if callable(v):
                value = hash_names.get(f"format_fn[{k}]")
            elif isinstance(v, str):
                value = v.strip()
            else:
                value = None
            if value in columns.keys():
                _warn(f"{elt_name}: format_fn[{k}] cannot be set to an column name '{value}'.")
            elif value:
                col_desc["formatFn"] = value
        else:
            _warn(f"{elt_name}: '{k}' is not a displayed column in format_fn[].")
    editable = attributes.get("editable", False)
    loveable = _is_boolean(editable) and _is_true(editable)
    loves = _get_name_indexed_property(attributes, "lov")
    for k, v in loves.items():  # pragma: no cover
        col_desc = _get_column_desc(columns, k)
        if col_desc and (
            loveable
            or not col_desc.get("notEditable", True)
            or t.cast(str, col_desc.get("type", "")).startswith("bool")
        ):
            value = v.strip().split(";") if isinstance(v, str) else v  # type: ignore[assignment]
            if value is not None and not isinstance(value, (list, tuple)):
                _warn(f"{elt_name}: lov[{k}] should be a list.")
                value = None
            if value is not None:
                new_value = list(filter(lambda i: i is not None, value))
                if len(new_value) < len(value):
                    col_desc["freeLov"] = True
                    value = new_value
                col_desc["lov"] = value
        elif not col_desc:
            _warn(f"{elt_name}: '{k}' is not a displayed column in lov[].")
    return columns



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import contextlib
import logging
import os
import pathlib
import re
import sys
import time
import typing as t
import webbrowser
from importlib import util
from random import choices, randint

from flask import (
    Blueprint,
    Flask,
    json,
    jsonify,
    make_response,
    render_template,
    render_template_string,
    request,
    send_from_directory,
)
from flask_cors import CORS
from flask_socketio import SocketIO
from gitignore_parser import parse_gitignore
from kthread import KThread
from werkzeug.serving import is_running_from_reloader

import __main__
from taipy.common.logger._taipy_logger import _TaipyLogger

from ._renderers.json import _TaipyJsonProvider
from .config import ServerConfig
from .custom._page import _ExternalResourceHandlerManager
from .utils import _is_in_notebook, _is_port_open, _RuntimeManager
from .utils._css import get_style

if t.TYPE_CHECKING:
    from .gui import Gui


class _Server:
    __RE_OPENING_CURLY = re.compile(r"([^\"])(\{)")
    __RE_CLOSING_CURLY = re.compile(r"(\})([^\"])")
    __OPENING_CURLY = r"\1&#x7B;"
    __CLOSING_CURLY = r"&#x7D;\2"
    _RESOURCE_HANDLER_ARG = "tprh"

    def __init__(
        self,
        gui: Gui,
        flask: t.Optional[Flask] = None,
        path_mapping: t.Optional[dict] = None,
        async_mode: t.Optional[str] = None,
        allow_upgrades: bool = True,
        server_config: t.Optional[ServerConfig] = None,
    ):
        self._gui = gui
        server_config = server_config or {}
        self._flask = flask
        if self._flask is None:
            flask_config: t.Dict[str, t.Any] = {"import_name": "Taipy"}
            if "flask" in server_config and isinstance(server_config["flask"], dict):
                flask_config.update(server_config["flask"])
            self._flask = Flask(**flask_config)
        if "SECRET_KEY" not in self._flask.config or not self._flask.config["SECRET_KEY"]:
            self._flask.config["SECRET_KEY"] = "TaIpY"

        # setup cors
        if "cors" not in server_config or (
            "cors" in server_config and (isinstance(server_config["cors"], dict) or server_config["cors"] is True)
        ):
            cors_config = (
                server_config["cors"] if "cors" in server_config and isinstance(server_config["cors"], dict) else {}
            )
            CORS(self._flask, **cors_config)

        # setup socketio
        socketio_config: t.Dict[str, t.Any] = {
            "cors_allowed_origins": "*",
            "ping_timeout": 10,
            "ping_interval": 5,
            "json": json,
            "async_mode": async_mode,
            "allow_upgrades": allow_upgrades,
        }
        if "socketio" in server_config and isinstance(server_config["socketio"], dict):
            socketio_config.update(server_config["socketio"])
        self._ws = SocketIO(self._flask, **socketio_config)

        self._apply_patch()

        # set json encoder (for Taipy specific types)
        self._flask.json_provider_class = _TaipyJsonProvider
        self._flask.json = self._flask.json_provider_class(self._flask)  # type: ignore

        self.__path_mapping = path_mapping or {}
        self.__ssl_context = server_config.get("ssl_context", None)
        self._is_running = False

        # Websocket (handle json message)
        # adding args for the one call with a server ack request
        @self._ws.on("message")
        def handle_message(message, *args) -> None:
            if "status" in message:
                _TaipyLogger._get_logger().info(message["status"])
            elif "type" in message:
                gui._manage_message(message["type"], message)

        @self._ws.on("connect")
        def handle_connect():
            gui._handle_connect()

        @self._ws.on("disconnect")
        def handle_disconnect():
            gui._handle_disconnect()

    def __is_ignored(self, file_path: str) -> bool:
        if not hasattr(self, "_ignore_matches"):
            __IGNORE_FILE = ".taipyignore"
            ignore_file = (
                (pathlib.Path(__main__.__file__).parent / __IGNORE_FILE) if hasattr(__main__, "__file__") else None
            )
            if not ignore_file or not ignore_file.is_file():
                ignore_file = pathlib.Path(self._gui._root_dir) / __IGNORE_FILE
            self._ignore_matches = (
                parse_gitignore(ignore_file) if ignore_file.is_file() and os.access(ignore_file, os.R_OK) else None
            )

        if callable(self._ignore_matches):
            return self._ignore_matches(file_path)
        return False

    def _get_default_blueprint(
        self,
        static_folder: str,
        template_folder: str,
        title: str,
        favicon: str,
        root_margin: str,
        scripts: t.List[str],
        styles: t.List[str],
        version: str,
        client_config: t.Dict[str, t.Any],
        watermark: t.Optional[str],
        css_vars: str,
        base_url: str,
    ) -> Blueprint:
        taipy_bp = Blueprint("Taipy", __name__, static_folder=static_folder, template_folder=template_folder)
        # Serve static react build

        @taipy_bp.route("/", defaults={"path": ""})
        @taipy_bp.route("/<path:path>")
        def my_index(path):
            resource_handler_id = request.cookies.get(_Server._RESOURCE_HANDLER_ARG, None)
            if resource_handler_id is not None:
                resource_handler = _ExternalResourceHandlerManager().get(resource_handler_id)
                if resource_handler is None:
                    reload_html = "<html><head><style>body {background-color: black; margin: 0;}</style></head><body><script>location.reload();</script></body></html>"  # noqa: E501
                    response = make_response(render_template_string(reload_html), 400)
                    response.set_cookie(
                        _Server._RESOURCE_HANDLER_ARG, "", secure=request.is_secure, httponly=True, expires=0, path="/"
                    )
                    return response
                try:
                    return resource_handler.get_resources(path, static_folder, base_url)
                except Exception as e:
                    raise RuntimeError("Can't get resources from custom resource handler") from e
            if path == "" or path == "index.html" or "." not in path:
                try:
                    return render_template(
                        "index.html",
                        title=title,
                        favicon=f"{favicon}?version={version}",
                        root_margin=root_margin,
                        watermark=watermark,
                        config=client_config,
                        scripts=scripts,
                        styles=styles,
                        version=version,
                        css_vars=css_vars,
                        base_url=base_url,
                    )
                except Exception:
                    raise RuntimeError(
                        "Something is wrong with the taipy-gui front-end installation. Check that the js bundle has been properly built (is Node.js installed?)."  # noqa: E501
                    ) from None

            if path == "taipy.status.json":
                return self._direct_render_json(self._gui._serve_status(pathlib.Path(template_folder) / path))
            if (file_path := str(os.path.normpath((base_path := static_folder + os.path.sep) + path))).startswith(
                base_path
            ) and os.path.isfile(file_path):
                return send_from_directory(base_path, path)
            # use the path mapping to detect and find resources
            for k, v in self.__path_mapping.items():
                if (
                    path.startswith(f"{k}/")
                    and (
                        file_path := str(os.path.normpath((base_path := v + os.path.sep) + path[len(k) + 1 :]))
                    ).startswith(base_path)
                    and os.path.isfile(file_path)
                ):
                    return send_from_directory(base_path, path[len(k) + 1 :])
            if (
                hasattr(__main__, "__file__")
                and (
                    file_path := str(
                        os.path.normpath((base_path := os.path.dirname(__main__.__file__) + os.path.sep) + path)
                    )
                ).startswith(base_path)
                and os.path.isfile(file_path)
                and not self.__is_ignored(file_path)
            ):
                return send_from_directory(base_path, path)
            if (
                (
                    file_path := str(os.path.normpath((base_path := self._gui._root_dir + os.path.sep) + path))
                ).startswith(base_path)
                and os.path.isfile(file_path)
                and not self.__is_ignored(file_path)
            ):
                return send_from_directory(base_path, path)
            return ("", 404)

        return taipy_bp

    # Update to render as JSX
    def _render(self, html_fragment, style, head, context):
        template_str = _Server.__RE_OPENING_CURLY.sub(_Server.__OPENING_CURLY, html_fragment)
        template_str = _Server.__RE_CLOSING_CURLY.sub(_Server.__CLOSING_CURLY, template_str)
        template_str = template_str.replace('"{!', "{")
        template_str = template_str.replace('!}"', "}")
        style = get_style(style)
        return self._direct_render_json(
            {
                "jsx": template_str,
                "style": (style + os.linesep) if style else "",
                "head": head or [],
                "context": context or self._gui._get_default_module_name(),
            }
        )

    def _direct_render_json(self, data):
        return jsonify(data)

    def get_flask(self):
        return self._flask

    def get_port(self):
        return self._port

    def test_client(self):
        return t.cast(Flask, self._flask).test_client()

    def _run_notebook(self):
        self._is_running = True
        self._ws.run(self._flask, host=self._host, port=self._port, debug=False, use_reloader=False)

    def _get_async_mode(self) -> str:
        return self._ws.async_mode  # type: ignore[reportAttributeAccessIssue]

    def _apply_patch(self):
        if self._get_async_mode() == "gevent" and util.find_spec("gevent"):
            from gevent import get_hub, monkey

            get_hub().NOT_ERROR += (KeyboardInterrupt,)
            if not monkey.is_module_patched("time"):
                monkey.patch_time()
        if self._get_async_mode() == "eventlet" and util.find_spec("eventlet"):
            from eventlet import monkey_patch, patcher  # type: ignore[reportMissingImport]

            if not patcher.is_monkey_patched("time"):
                monkey_patch(time=True)

    def _get_random_port(
        self, port_auto_ranges: t.Optional[t.List[t.Union[int, t.Tuple[int, int]]]] = None
    ):  # pragma: no cover
        port_auto_ranges = port_auto_ranges or [(49152, 65535)]
        random_weights = [1 if isinstance(r, int) else abs(r[1] - r[0]) + 1 for r in port_auto_ranges]
        while True:
            random_choices = [
                r if isinstance(r, int) else randint(min(r[0], r[1]), max(r[0], r[1])) for r in port_auto_ranges
            ]
            port = choices(random_choices, weights=random_weights)[0]
            if port not in _RuntimeManager().get_used_port() and not _is_port_open(self._host, port):
                return port

    def run(
        self,
        host,
        port,
        client_url,
        debug,
        use_reloader,
        flask_log,
        run_in_thread,
        allow_unsafe_werkzeug,
        notebook_proxy,
        port_auto_ranges,
    ):
        host_value = host if host != "0.0.0.0" else "localhost"
        self._host = host
        if port == "auto":
            port = self._get_random_port(port_auto_ranges)
        server_url = f"http://{host_value}:{port}"
        self._port = port
        if _is_in_notebook() and notebook_proxy:  # pragma: no cover
            from .utils.proxy import NotebookProxy

            # Start proxy if not already started
            self._proxy = NotebookProxy(gui=self._gui, listening_port=port)
            self._proxy.run()
            self._port = self._get_random_port()
        if _is_in_notebook() or run_in_thread:
            runtime_manager = _RuntimeManager()
            runtime_manager.add_gui(self._gui, port)
        if debug and not is_running_from_reloader() and _is_port_open(host_value, port):
            raise ConnectionError(
                f"Port {port} is already opened on {host} because another application is running on the same port.\nPlease pick another port number and rerun with the 'port=<new_port>' setting.\nYou can also let Taipy choose a port number for you by running with the 'port=\"auto\"' setting."  # noqa: E501
            )
        if not flask_log:
            log = logging.getLogger("werkzeug")
            log.disabled = True
            if not is_running_from_reloader():
                _TaipyLogger._get_logger().info(f" * Server starting on {server_url}")
            else:
                _TaipyLogger._get_logger().info(f" * Server reloaded on {server_url}")
            if client_url is not None:
                client_url = client_url.format(port=port)
                _TaipyLogger._get_logger().info(f" * Application is accessible at {client_url}")
        if not is_running_from_reloader() and self._gui._get_config("run_browser", False):
            webbrowser.open(client_url or server_url, new=2)
        if _is_in_notebook() or run_in_thread:
            self._thread = KThread(target=self._run_notebook)
            self._thread.start()
            return
        self._is_running = True
        run_config = {
            "app": self._flask,
            "host": host,
            "port": port,
            "debug": debug,
            "use_reloader": use_reloader,
        }
        if self.__ssl_context is not None:
            run_config["ssl_context"] = self.__ssl_context
        # flask-socketio specific conditions for 'allow_unsafe_werkzeug' parameters to be popped out of kwargs
        if self._get_async_mode() == "threading" and (not sys.stdin or not sys.stdin.isatty()):
            run_config = {**run_config, "allow_unsafe_werkzeug": allow_unsafe_werkzeug}
        try:
            self._ws.run(**run_config)
        except KeyboardInterrupt:
            pass

    def stop_thread(self):
        if hasattr(self, "_thread") and self._thread.is_alive() and self._is_running:
            self._is_running = False
            with contextlib.suppress(Exception):
                if self._get_async_mode() == "gevent":
                    if self._ws.wsgi_server is not None:  # type: ignore[reportAttributeAccessIssue]
                        self._ws.wsgi_server.stop()  # type: ignore[reportAttributeAccessIssue]
                    else:
                        self._thread.kill()
                else:
                    self._thread.kill()
            while _is_port_open(self._host, self._port):
                time.sleep(0.1)

    def stop_proxy(self):
        if hasattr(self, "_proxy"):
            self._proxy.stop()



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t

import numpy as np

from .base import Decimator


class ScatterDecimator(Decimator):
    """A decimator designed for scatter charts.

    This algorithm fits the data points into a grid. If multiple points are in the same grid
    cell, depending on the chart configuration, some points are removed to reduce the number
    points being displayed.

    This class can only be used with scatter charts.
    """

    _CHART_MODES = ["markers"]

    def __init__(
        self,
        binning_ratio: t.Optional[float] = None,
        max_overlap_points: t.Optional[int] = None,
        threshold: t.Optional[int] = None,
        zoom: t.Optional[bool] = True,
        # on_decimate: t.Optional[t.Callable] = None,
        # apply_decimator: t.Optional[t.Callable] = None,
    ):
        """Initialize a new `ScatterDecimator`.

        Arguments:
            binning_ratio (Optional[float]): the size of the data grid for the algorithm.
                The higher the value, the smaller the grid size.
            max_overlap_points (Optional(int)): the maximum number of points for a single cell
                within the data grid within the algorithm. This dictates how dense a single
                grid cell can be.
            threshold (Optional[int]): The minimum amount of data points before the
                decimation is applied.
            zoom (Optional[bool]): set to True to reapply the decimation
                when zoom or re-layout events are triggered.
        """
        # on_decimate (Optional[Callable]): an user-defined function that is executed when the decimator
        #     is found during runtime. This function can be used to provide custom decimation logic.
        # apply_decimator (Optional[Callable]): an user-defined function that is executed when the decimator
        #     is applied to modify the data.
        super().__init__(threshold, zoom)
        binning_ratio = binning_ratio if binning_ratio is not None else 1
        self._binning_ratio = binning_ratio if binning_ratio > 0 else 1
        self._max_overlap_points = max_overlap_points if max_overlap_points is not None else 3

    def _decimate(self, data: np.ndarray, payload: t.Dict[str, t.Any]) -> np.ndarray:
        n_rows = data.shape[0]
        mask = np.empty(n_rows, dtype=bool)
        width = payload.get("width", None)
        height = payload.get("height", None)
        if width is None or height is None:
            mask.fill(True)
            return mask
        mask.fill(False)
        grid_x, grid_y = round(width / self._binning_ratio), round(height / self._binning_ratio)
        x_col, y_col = data[:, 0], data[:, 1]
        min_x: float = np.amin(x_col)
        max_x: float = np.amax(x_col)
        min_y: float = np.amin(y_col)
        max_y: float = np.amax(y_col)
        min_max_x_diff, min_max_y_diff = max_x - min_x, max_y - min_y
        x_grid_map = np.rint((x_col - min_x) * grid_x / min_max_x_diff).astype(int)
        y_grid_map = np.rint((y_col - min_y) * grid_y / min_max_y_diff).astype(int)
        z_grid_map = None
        grid_shape = (grid_x + 1, grid_y + 1)
        if len(data[0]) == 3:
            grid_z = grid_x
            grid_shape = (grid_x + 1, grid_y + 1, grid_z + 1)  # type: ignore
            z_col = data[:, 2]
            min_z: float = np.amin(z_col)
            max_z: float = np.amax(z_col)
            min_max_z_diff = max_z - min_z
            z_grid_map = np.rint((z_col - min_z) * grid_z / min_max_z_diff).astype(int)
        grid = np.empty(grid_shape, dtype=int)
        grid.fill(0)
        if z_grid_map is not None:
            for i in np.arange(n_rows):
                if grid[x_grid_map[i], y_grid_map[i], z_grid_map[i]] < self._max_overlap_points:
                    grid[x_grid_map[i], y_grid_map[i], z_grid_map[i]] += 1
                    mask[i] = True
        else:
            for i in np.arange(n_rows):
                if grid[x_grid_map[i], y_grid_map[i]] < self._max_overlap_points:
                    grid[x_grid_map[i], y_grid_map[i]] += 1
                    mask[i] = True
        return mask



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import inspect
import typing as t
from types import FrameType

from .utils import _filter_locals, _get_module_name_from_frame

if t.TYPE_CHECKING:
    from ._page import _Page
    from .gui import Gui


class Page:
    """Generic page generator.

    The `Page` class transforms template text into actual pages that can be displayed
    on a web browser.

    When a page is requested to be displayed, it is converted into HTML
    code that can be sent to the client. All control placeholders are
    replaced by their respective graphical component so you can show
    your application variables and interact with them.
    """

    page_type: str = "Taipy"

    def __init__(self, **kwargs) -> None:
        from .custom import Page as CustomPage

        self._class_module_name = ""
        self._class_locals: t.Dict[str, t.Any] = {}
        self._frame: t.Optional[FrameType] = None
        self._renderer: t.Optional[Page] = self.create_page()
        if "frame" in kwargs:
            self._frame = kwargs.get("frame")
        elif self._renderer:
            self._frame = self._renderer._frame
        elif isinstance(self, CustomPage):
            self._frame = t.cast(FrameType, t.cast(FrameType, inspect.stack()[2].frame))
            # Allow CustomPage class to be inherited
            if len(inspect.stack()) > 3 and inspect.stack()[2].function != "<module>":
                self._frame = t.cast(FrameType, t.cast(FrameType, inspect.stack()[3].frame))
        elif len(inspect.stack()) < 4:
            raise RuntimeError(f"Can't resolve module. Page '{type(self).__name__}' is not registered.")
        else:
            self._frame = t.cast(FrameType, t.cast(FrameType, inspect.stack()[3].frame))
        if self._renderer:
            # Extract the page module's attributes and methods
            cls = type(self)
            cls_locals = dict(vars(self))
            functions = [
                i[0]
                for i in inspect.getmembers(cls)
                if not i[0].startswith("_") and inspect.isroutine(i[1])
            ]
            for f in functions:
                func = getattr(self, f)
                if hasattr(func, "__func__") and func.__func__ is not None:
                    cls_locals[f] = func.__func__
            self._class_module_name = cls.__name__
            self._class_locals = cls_locals
        # Special variables only use for page reloading in notebook context
        self._notebook_gui: t.Optional["Gui"] = None
        self._notebook_page: t.Optional["_Page"] = None
        self.set_style(t.cast(dict, kwargs.get("style", None)))

    def create_page(self) -> t.Optional[Page]:
        """Create the page content for page modules.

        If this page is a page module, this method must be overloaded and return the page content.

        This method should never be called directly: only the Taipy GUI internals will.

        The default implementation returns None, indicating that this class does not implement
        a page module.

        Returns:
            The page content for this Page subclass, making it a page module.
        """
        return None

    def set_content(self, content: str) -> None:
        """Set a new page content.

        Reads the new page content and re-initializes the `Page^` instance to reflect the change.

        !!! important
            This function can only be used in an IPython notebook context.

        Arguments:
            content (str): The text content or the path to the file holding the text to be transformed.
                If *content* is a path to a readable file, the file is read entirely as the text
                template.

        Exceptions:
            RuntimeError: If this method is called outside an IPython notebook context.
        """
        # Implemented in the private _Renderer class
        raise NotImplementedError("Not in a valid Page class.")

    def _get_locals(self) -> t.Optional[t.Dict[str, t.Any]]:
        return (
            self._class_locals
            if self._is_class_module()
            else None
            if (frame := self._get_frame()) is None
            else _filter_locals(frame.f_locals)
        )

    def _is_class_module(self):
        return self._class_module_name != ""

    def _get_frame(self):
        if not hasattr(self, "_frame"):
            raise RuntimeError(f"Page '{type(self).__name__}' was not registered correctly.")
        return self._frame

    def _get_module_name(self) -> t.Optional[str]:
        return (
            None
            if (frame := self._get_frame()) is None
            else f"{_get_module_name_from_frame(frame)}{'.' if self._class_module_name else ''}{self._class_module_name}"  # noqa: E501
        )

    def _get_content_detail(self, gui) -> str:
        return f"in class {type(self).__name__}"

    def render(self, gui) -> str:
        if self._renderer is not None:
            return self._renderer.render(gui)
        return "<h1>No renderer found for page</h1>"

    def set_style(self, style: t.Dict[str, t.Dict[str, t.Any]]) -> Page:
        """Set the style for this page.

        The *style* parameter must contain a series of CSS rules that apply to the generated
        page.<br/>
        Each key of this dictionary should be a CSS selector and its associated value must be
        a CSS declaration or a CSS rule itself, benefiting from
        [nested CSS](https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_nesting/Using_CSS_nesting)
        features.

        For example, you could set the *style* parameter to:
        ```python
        {
          "class1": {
            "css_property1": "css_value1",
          }
          "class2": {
            "class3": {
              "css_property2": "css_value2",
            }
          }
        }
        ```
        That would set the "css_property1" to "css_value1" for all elements with the "class1"
        class, and "css_property2" to "css_value2" for all elements with the "class3" class that
        are descendants of elements with the "class2" class.

        Arguments:
            style (dict): A dictionary describing the style as CSS or Nested CSS.

        Returns:
            This `Page` instance.
        """
        self.__style = style if isinstance(style, dict) else None
        return self

    def _get_style(self):
        return self.__style



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from .config import Config, Stylekit

_default_stylekit: Stylekit = {
    # Primary and secondary colors
    "color_primary": "#ff6049",
    "color_secondary": "#293ee7",
    # Contextual color
    "color_error": "#FF595E",
    "color_warning": "#FAA916",
    "color_success": "#96E6B3",
    # Background and elevation color for LIGHT MODE
    "color_background_light": "#f0f5f7",
    "color_paper_light": "#ffffff",
    # Background and elevation color for DARK MODE
    "color_background_dark": "#152335",
    "color_paper_dark": "#1f2f44",
    # DEFINING FONTS
    # Set main font family
    "font_family": "Lato, Arial, sans-serif",
    # DEFINING ROOT STYLES
    # Set root margin
    "root_margin": "1rem",
    # DEFINING SHAPES
    # Base border radius in px
    "border_radius": 8,
    # DEFINING MUI COMPONENTS STYLES
    # Matching input and button height in css size unit
    "input_button_height": "48px",
}

# Default config loaded by app.py
default_config: Config = {
    "allow_unsafe_werkzeug": False,
    "async_mode": "gevent",
    "change_delay": None,
    "chart_dark_template": None,
    "base_url": "/",
    "client_url": None,
    "dark_mode": True,
    "dark_theme": None,
    "debug": False,
    "extended_status": False,
    "favicon": None,
    "flask_log": False,
    "host": "127.0.0.1",
    "light_theme": None,
    "margin": "1em",
    "ngrok_token": "",
    "notebook_proxy": True,
    "notification_duration": 3000,
    "port": 5000,
    "port_auto_ranges": [(49152, 65535)],
    "propagate": True,
    "run_browser": True,
    "run_in_thread": False,
    "run_server": True,
    "server_config": None,
    "single_client": False,
    "state_retention_period": 0,
    "system_notification": False,
    "theme": None,
    "time_zone": None,
    "title": None,
    "stylekit": _default_stylekit.copy(),
    "upload_folder": None,
    "use_arrow": False,
    "use_reloader": False,
    "watermark": "Taipy inside",
    "webapp_path": None,
}



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd

from ..._warnings import _warn


class Decimator(ABC):
    """Base class for decimating chart data.

    *Decimating* is the term used to name the process of reducing the number of
    data points displayed in charts while retaining the overall shape of the traces.
    `Decimator` is a base class that does decimation on data sets.

    Taipy GUI comes out-of-the-box with several implementation of this class for
    different use cases.
    """

    _CHART_MODES: t.List[str] = []

    def __init__(
        self,
        threshold: t.Optional[int],
        zoom: t.Optional[bool],
        # apply_decimator: t.Optional[t.Callable] = None,
        # on_decimate: t.Optional[t.Callable] = None,
    ) -> None:  # noqa: E501
        """Initialize a new `Decimator`.

        Arguments:
            threshold (Optional[int]): The minimum amount of data points before the
                decimator class is applied.
            zoom (Optional[bool]): set to True to reapply the decimation
                when zoom or re-layout events are triggered.
        """
        # on_decimate (Optional[Callable]): an user-defined function that is executed when the decimator
        #     is found during runtime. This function can be used to provide custom decimation logic.
        # apply_decimator (Optional[Callable]): an user-defined function that is executed when the decimator
        #     is applied to modify the data.
        super().__init__()
        self.threshold = threshold
        self._zoom = zoom if zoom is not None else True
        self.__user_defined_on_decimate = None
        self.__user_defined_apply_decimator = None

    def _is_applicable(self, data: t.Any, nb_rows_max: int, chart_mode: str):
        if chart_mode not in self._CHART_MODES:
            _warn(
                f"Decimator '{type(self).__name__}' is not optimized for chart mode '{chart_mode}'. Consider using other chart mode such as '{f'{chr(39)}, {chr(39)}'.join(self._CHART_MODES)}.'"  # noqa: E501
            )
        if self.threshold is None:
            if nb_rows_max < len(data):
                return True
        elif self.threshold < len(data):
            return True
        return False

    def __get_indexed_df_col(self, df):
        index = 0
        while f"tAiPy_index_{index}" in df.columns:
            index += 1
        result = f"tAiPy_index_{index}"
        df[result] = df.index
        return result

    def _df_relayout(
        self,
        dataframe: pd.DataFrame,
        x_column: t.Optional[str],
        y_column: str,
        chart_mode: str,
        x0: t.Optional[float],
        x1: t.Optional[float],
        y0: t.Optional[float],
        y1: t.Optional[float],
        is_copied: bool,
    ):
        if chart_mode not in ["lines+markers", "lines", "markers"]:
            _warn(
                f"Decimator zoom feature is not applicable for '{chart_mode}' chart_mode. It is only applicable for 'lines+markers', 'lines', and 'markers' chart modes."  # noqa: E501
            )
            return dataframe, is_copied
        # if chart data is invalid
        if x0 is None and x1 is None and y0 is None and y1 is None:
            return dataframe, is_copied
        df = dataframe if is_copied else dataframe.copy()
        is_copied = True
        has_x_col = True

        if not x_column:
            x_column = self.__get_indexed_df_col(df)
            has_x_col = False

        df_filter_conditions = []
        # filter by x column by default
        if x0 is not None:
            df_filter_conditions.append(df[x_column] > x0)
        if x1 is not None:
            df_filter_conditions.append(df[x_column] < x1)
        # y column will be filtered only if chart_mode is not lines+markers (eg. markers)
        if chart_mode not in ["lines+markers", "lines"]:
            if y0 is not None:
                df_filter_conditions.append(df[y_column] > y0)
            if y1 is not None:
                df_filter_conditions.append(df[y_column] < y1)
        if df_filter_conditions:
            df = df.loc[np.bitwise_and.reduce(df_filter_conditions)]
        if not has_x_col:
            df.drop(x_column, axis=1, inplace=True)
        return df, is_copied

    def _df_apply_decimator(
        self,
        dataframe: pd.DataFrame,
        x_column_name: t.Optional[str],
        y_column_name: str,
        z_column_name: str,
        payload: t.Dict[str, t.Any],
        is_copied: bool,
    ):
        df = dataframe if is_copied else dataframe.copy()
        if not x_column_name:
            x_column_name = self.__get_indexed_df_col(df)
        column_list = [x_column_name, y_column_name, z_column_name] if z_column_name else [x_column_name, y_column_name]
        points = df[column_list].to_numpy()
        mask = self._decimate(points, payload)
        return df[mask], is_copied

    def _on_decimate_df(
        self,
        df: pd.DataFrame,
        decimator_instance_payload: t.Dict[str, t.Any],
        decimator_payload: t.Dict[str, t.Any],
        is_copied: bool = False,
        filter_unused_columns: bool = True,
    ):
        decimator_var_name = decimator_instance_payload.get("decimator")
        x_column, y_column, z_column = (
            decimator_instance_payload.get("xAxis", ""),
            decimator_instance_payload.get("yAxis", ""),
            decimator_instance_payload.get("zAxis", ""),
        )
        chart_mode = decimator_instance_payload.get("chartMode", "")
        if self._zoom and "relayoutData" in decimator_payload is not None and not z_column:
            relayout_data = decimator_payload.get("relayoutData", {})
            x0 = relayout_data.get("xaxis.range[0]")
            x1 = relayout_data.get("xaxis.range[1]")
            y0 = relayout_data.get("yaxis.range[0]")
            y1 = relayout_data.get("yaxis.range[1]")

            df, is_copied = self._df_relayout(
                t.cast(pd.DataFrame, df), x_column, y_column, chart_mode, x0, x1, y0, y1, is_copied
            )

        nb_rows_max = decimator_payload.get("width")
        is_decimator_applied = False
        if nb_rows_max and self._is_applicable(df, nb_rows_max, chart_mode):
            try:
                df, is_copied = self._apply_decimator(
                    t.cast(pd.DataFrame, df),
                    x_column,
                    y_column,
                    z_column,
                    payload=decimator_payload,
                    is_copied=is_copied,
                )
                is_decimator_applied = True
            except Exception as e:
                _warn(f"Limit rows error with {decimator_var_name} for Dataframe", e)
        # only include columns involving the decimator
        if filter_unused_columns:
            filterd_columns = [x_column, y_column, z_column] if z_column else [x_column, y_column]
            df = df.filter(filterd_columns, axis=1)
        return df, is_decimator_applied, is_copied

    def _on_decimate(
        self,
        df: pd.DataFrame,
        decimator_instance_payload: t.Dict[str, t.Any],
        decimator_payload: t.Dict[str, t.Any],
        is_copied: bool = False,
        filter_unused_columns: bool = True,
    ) -> t.Tuple:
        """NOT DOCUMENTED
        This function is executed whenever a decimator is found during runtime.

        Users can override this function by providing an alternate implementation inside the constructor
        to provide custom decimation logic.

        Arguments:
            df (pandas.DataFrame): The DataFrame that will be decimated.
            decimator_instance_payload (Dict[str, any]): The payload for the current instance of decimator. Each
                decimation request might contain multiple decimators to handle multiple traces.
            decimator_payload (Dict[str, any]): The full decimator payload including the current decimator.
            is_copied (bool): A flag to indicate if the DataFrame is copied.
            filter_unused_columns (bool): A flag to indicate if the DataFrame columns should be filtered to only
                include the columns that are involved with the decimator.

        Returns:
            A tuple containing the decimated DataFrame, a flag indicating if the decimator is applied,
            and a flag indicating if the DataFrame was copied.
        """
        if self.__user_defined_on_decimate and callable(self.__user_defined_on_decimate):
            try:
                return self.__user_defined_on_decimate(
                    df, decimator_instance_payload, decimator_payload, is_copied, filter_unused_columns
                )
            except Exception as e:
                _warn("Error executing user defined on_decimate function: ", e)
        return self._on_decimate_df(df, decimator_instance_payload, decimator_payload, filter_unused_columns)

    def _apply_decimator(
        self,
        dataframe: pd.DataFrame,
        x_column_name: t.Optional[str],
        y_column_name: str,
        z_column_name: str,
        payload: t.Dict[str, t.Any],
        is_copied: bool,
    ) -> t.Tuple:
        """NOT DOCUMENTED
        This function is executed whenever a decimator is applied to the data.

        This function is used by default the `on_decimate` function.
        Users can override this function by providing an alternate function inside the constructor
        to provide custom handling only when the decimator is applied. This avoids the need to override
        the default `on_decimate` handling logic.

        Arguments:
            dataframe (pandas.DataFrame): The DataFrame that will be decimated.
            x_column_name (Optional[str]): The name of the x-axis column.
            y_column_name (str): The name of the y-axis column.
            z_column_name (str): The name of the z-axis column.
            payload (Dict[str, any]): The payload for the current decimator.
            is_copied (bool): A flag to indicate if the DataFrame is copied.

        Returns:
            A tuple containing the decimated DataFrame and a flag indicating if the DataFrame was copied
        """
        if self.__user_defined_apply_decimator and callable(self.__user_defined_apply_decimator):
            try:
                return self.__user_defined_apply_decimator(
                    dataframe, x_column_name, y_column_name, z_column_name, payload, is_copied
                )
            except Exception as e:
                _warn("Error executing user defined apply_decimator function: ", e)
        return self._df_apply_decimator(dataframe, x_column_name, y_column_name, z_column_name, payload, is_copied)

    @abstractmethod
    def _decimate(self, data: np.ndarray, payload: t.Dict[str, t.Any]) -> np.ndarray:
        """NOT DOCUMENTED
        Decimate the dataset.

        This method is executed when the appropriate conditions specified in the
        constructor are met. This function implements the algorithm that determines
        which data points are kept or dropped.

        Arguments:
            data (numpy.array): An array containing all the data points represented as
                tuples.
            payload (Dict[str, any]): additional information on charts that is provided
                at runtime.

        Returns:
            An array of Boolean mask values. The array should set True or False for each
                of its indexes where True indicates that the corresponding data point
                from *data* should be preserved, or False requires that this
                data point be dropped.
        """
        raise NotImplementedError



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import typing as t
from types import SimpleNamespace

from .._warnings import _warn

if t.TYPE_CHECKING:
    from ..gui import Gui


class _DataScopes:
    _GLOBAL_ID = "global"
    _META_PRE_RENDER = "pre_render"
    _DEFAULT_METADATA = {_META_PRE_RENDER: False}

    def __init__(self, gui: "Gui") -> None:
        self.__gui = gui
        self.__scopes: t.Dict[str, SimpleNamespace] = {_DataScopes._GLOBAL_ID: SimpleNamespace()}
        # { scope_name: { metadata: value } }
        self.__scopes_metadata: t.Dict[str, t.Dict[str, t.Any]] = {
            _DataScopes._GLOBAL_ID: _DataScopes._DEFAULT_METADATA.copy()
        }
        self.__single_client = True

    def set_single_client(self, value: bool) -> None:
        self.__single_client = value

    def is_single_client(self) -> bool:
        return self.__single_client

    def get_scope(self, client_id: t.Optional[str]) -> t.Tuple[SimpleNamespace, t.Dict[str, str]]:
        if self.__single_client:
            return self.__scopes[_DataScopes._GLOBAL_ID], self.__scopes_metadata[_DataScopes._GLOBAL_ID]
        # global context in case request is not registered or client_id is not
        # available (such as in the context of running tests)
        if not client_id:
            _warn("Empty session id, using global scope instead.")
            return self.__scopes[_DataScopes._GLOBAL_ID], self.__scopes_metadata[_DataScopes._GLOBAL_ID]
        if client_id not in self.__scopes:
            _warn(
                f"Session id {client_id} not found in data scope. Taipy will automatically create a scope for this session id but you may have to reload your page."  # noqa: E501
            )
            self.create_scope(client_id)
        return self.__scopes[client_id], self.__scopes_metadata[client_id]

    def get_all_scopes(self) -> t.Dict[str, SimpleNamespace]:
        return self.__scopes

    def create_scope(self, id: str) -> None:
        if self.__single_client:
            return
        if id is None:
            _warn("Empty session id, might be due to unestablished WebSocket connection.")
            return
        if id not in self.__scopes:
            self.__scopes[id] = SimpleNamespace()
            self.__scopes_metadata[id] = _DataScopes._DEFAULT_METADATA.copy()
            # Propagate shared variables to the new scope from the global scope
            for var in self.__gui._get_shared_variables():
                if hasattr(self.__scopes[_DataScopes._GLOBAL_ID], var):
                    setattr(self.__scopes[id], var, getattr(self.__scopes[_DataScopes._GLOBAL_ID], var, None))

    def delete_scope(self, id: str) -> None:  # pragma: no cover
        if self.__single_client:
            return
        if id is None:
            _warn("Empty session id, might be due to unestablished WebSocket connection.")
            return
        if id in self.__scopes:
            del self.__scopes[id]
            del self.__scopes_metadata[id]



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import contextlib
import typing as t
import warnings
from threading import Thread
from urllib.parse import quote as urlquote
from urllib.parse import urlparse

from twisted.internet import reactor
from twisted.web.proxy import ProxyClient, ProxyClientFactory
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET, Site

from .is_port_open import _is_port_open

# flake8: noqa: E402
from .singleton import _Singleton

warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message="You do not have a working installation of the service_identity module: 'No module named 'service_identity''.*",  # noqa: E501
)


if t.TYPE_CHECKING:
    from ..gui import Gui


def _modifiedHandleResponseEnd(self):
    if self._finished:
        return
    self._finished = True
    with contextlib.suppress(Exception):
        self.father.finish()
    self.transport.loseConnection()


setattr(ProxyClient, "handleResponseEnd", _modifiedHandleResponseEnd)  # noqa: B010


class _TaipyReverseProxyResource(Resource):
    proxyClientFactoryClass = ProxyClientFactory

    def __init__(self, host, path, gui: "Gui", reactor=reactor):
        Resource.__init__(self)
        self.host = host
        self.path = path
        self.reactor = reactor
        self._gui = gui

    def getChild(self, path, request):
        return _TaipyReverseProxyResource(
            self.host,
            self.path + b"/" + urlquote(path, safe=b"").encode("utf-8"),
            self._gui,
            self.reactor,
        )

    def _get_port(self):
        return self._gui._server._port

    def render(self, request):
        port = self._get_port()
        host = self.host if port == 80 else "%s:%d" % (self.host, port)
        request.requestHeaders.setRawHeaders(b"host", [host.encode("ascii")])
        request.content.seek(0, 0)
        rest = self.path + b"?" + qs if (qs := urlparse(request.uri)[4]) else self.path
        clientFactory = self.proxyClientFactoryClass(
            request.method,
            rest,
            request.clientproto,
            request.getAllHeaders(),
            request.content.read(),
            request,
        )
        self.reactor.connectTCP(self.host, port, clientFactory)
        return NOT_DONE_YET


class NotebookProxy(object, metaclass=_Singleton):
    def __init__(self, gui: "Gui", listening_port: int) -> None:
        self._listening_port = listening_port
        self._gui = gui
        self._is_running = False

    def run(self):
        if self._is_running:
            return
        host = self._gui._get_config("host", "127.0.0.1")
        port = self._listening_port
        if _is_port_open(host, port):
            raise ConnectionError(
                f"Port {port} is already opened on {host}. You have another server application running on the same port."  # noqa: E501
            )
        site = Site(_TaipyReverseProxyResource(host, b"", self._gui))
        reactor.listenTCP(port, site)
        Thread(target=reactor.run, args=(False,)).start()
        self._is_running = True

    def stop(self):
        if not self._is_running:
            return
        self._is_running = False
        reactor.stop()



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

_replace_dict = {".": "__", "[": "_SqrOp_", "]": "_SqrCl_"}


def _get_client_var_name(var_name: str) -> str:
    for k, v in _replace_dict.items():
        var_name = var_name.replace(k, v)
    return var_name


def _to_camel_case(value: str, upcase_first=False) -> str:
    if not isinstance(value, str):
        raise Exception("_to_camel_case allows only string parameter")

    if len(value) <= 1:
        return value.lower()
    value = value.replace("_", " ").title().replace(" ", "").replace("[", "_").replace("]", "_")
    return value[0].lower() + value[1:] if not upcase_first else value


def _get_broadcast_var_name(s: str) -> str:
    return _get_client_var_name(f"_bc_{s}")



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import sys
import traceback
import typing as t
import warnings


class TaipyGuiWarning(UserWarning):
    """NOT DOCUMENTED

    Warning category for Taipy warnings generated in user code.
    """

    _tp_debug_mode = False

    @staticmethod
    def set_debug_mode(debug_mode: bool):
        TaipyGuiWarning._tp_debug_mode = (
            debug_mode if debug_mode else hasattr(sys, "gettrace") and sys.gettrace() is not None
        )


def _warn(message: str, e: t.Optional[BaseException] = None):
    warnings.warn(
        (
            f"{message}:\n{''.join(traceback.format_exception(type(e), e, e.__traceback__))}"
            if e and TaipyGuiWarning._tp_debug_mode
            else f"{message}:\n{e}"
            if e
            else message
        ),
        TaipyGuiWarning,
        stacklevel=2,
    )
